/*******************************************************************************
* Author: Martin Langer, Copyright: GPL
* This program is made to test the calculation-time with floatingpoint numbers 
* Chip type           : ATMEGA16
* Clock frequency     : External clock 7,372 Mhz
*******************************************************************************/
#include <stdlib.h>
#include <avr/io.h>
#include <inttypes.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include "uart.h"


/* define CPU frequency in Mhz here if not defined in Makefile */
#ifndef F_CPU
#define F_CPU 7372800UL
#endif


#define UART_BAUD_RATE      115200 

uint16_t ReadChannel(uint8_t mux_adc_channel);  /* ADC-Auslese-Funktion deklarieren */
float PI_Regler(float ek_0);						/* Funktion PI-Regler deklarieren */

float yk=0;
uint16_t wk=512;
uint16_t xk=0;
int16_t ek;

uint16_t treg1,treg2,tregdiff;
uint8_t zaehler=1;
volatile uint8_t ovzaehler=1;

char buffer[10];
char fbuffer[10];

ISR (SIG_OVERFLOW1)
{
ovzaehler=ovzaehler+1;
}
void
ioinit (void) /* Note [5] */
{
    
//	TCCR0 |= ( 1 << CS02 ); 	// Prescaler=256 d.h. Timer luft mit CPU-Takt/256
//	TCCR0 = ( 1 << CS00  ) | ( 1 << CS02  ); 	// Prescaler=1024 d.h. Timer luft mit CPU/1024-Takt
	TCCR1B |= ( 1 << CS12 ); 	// Prescaler=256 d.h. Timer luft mit CPU-Takt/256
//  TCCR2 |= ( 1 << CS20  ); 	// Prescaler=256 d.h. Timer luft mit CPU-Takt/1024
	
	uart_init( UART_BAUD_SELECT(UART_BAUD_RATE,F_CPU) ); 
	
	sei(); // Globale Interruptfreigabe
}

void main (void)
{
		  ioinit();
		  while (1){
			
			uart_puts("\f"); /* Bildschirm lschen */
			uart_puts("\r"); /* erste Position in neuer Zeile */
	
			uart_puts("Es wird ein AD-Wert an Kanal0 eingelesen. ");
			uart_puts("\n\r");   /* Zeilenumbruch */
			uart_puts("Daraus wird in  mittels eines PI-Regelalorithmus ein Stellwert berechnet.\n\r");
			uart_puts("\n\r");   /* Zeilenumbruch */
			
			xk = ReadChannel(0);    //  Channel 0 (Drehpoti) auslesen--> liefert Werte von 0 bis 1024
			
			utoa( xk , buffer, 10); /* convert interger into string */
			uart_puts("Der ADC-Wert an Kanal0 lautet: ");
			uart_puts(buffer);
			uart_puts("\n\r");   /* Zeilenumbruch */
			
				 
			TCNT1=0;
			TIMSK |= ( 1 << TOIE1 ); 				    // bei Timer1-berlauf wird Interrupt ausgelst
			treg1=TCNT0; // Zhlerstand1 sichern
			ek = wk-xk;
			yk = PI_Regler(ek);
			treg2=TCNT1; // Zhlerstand2 sichern
			TIMSK &= ~( 1 << TOIE1 ); 				    // bei Timer1-berlauf wird kein Interrupt ausgelst	
			
			itoa( ek , buffer, 10); /* convert interger into string */
			uart_puts("Die Ist-Sollwertdifferenz lautet: ");
			uart_puts(buffer);
			uart_puts("\n\r");   /* Zeilenumbruch */
		    
			dtostrf( yk , 6, 3, fbuffer); /* convert float into string */
			uart_puts("Die Regelerausgangsgroesse lautet: ");
			uart_puts(fbuffer);
			uart_puts("\n\r");   /* Zeilenumbruch */
			
			tregdiff=treg2-treg1;
			
			utoa( tregdiff , buffer, 10); /* convert interger into string */
			uart_puts("Die errechnete Timerdifferenz lautet: ");
			uart_puts(buffer);
			uart_puts("\n\r");   /* Zeilenumbruch */
		    
			
		    utoa( zaehler , buffer, 10); /* convert interger into string */
			uart_puts("Das ist der ");
			uart_puts(buffer);
			uart_puts(" Durchlauf");
			uart_puts("\n\r");   /* Zeilenumbruch */
		    
			utoa( ovzaehler , buffer, 10); /* convert interger into string */
			uart_puts("Das ist der ");
			uart_puts(buffer);
			uart_puts(" Timer-Overflow-Interrupt");
			uart_puts("\n\r");   /* Zeilenumbruch */
		    
			_delay_ms(100);
			_delay_ms(100);
			_delay_ms(100);
			_delay_ms(100);
			_delay_ms(100);
			_delay_ms(100);
			
			zaehler=zaehler+1;
			}
		  
}
uint16_t ReadChannel(uint8_t mux_adc_channel) //Unsere Funktion zum ADC-Channel aus lesen
{
  uint8_t i;
  uint16_t result = 0;         //Initialisieren wichtig, da lokale Variablen
                               //nicht automatisch initialisiert werden und
                               //zufllige Werte haben. Sonst kann Quatsch rauskommen
  ADCSRA = (1<<ADEN) | (1<<ADPS2) | (1<<ADPS0);   // Frequenzvorteiler auf 32 (Mega8 ((8 MHz)) setzen und ADC aktivieren 
// ADCSRA = (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1);    // Frequenzvorteiler auf 64 (Mega16(16 Mhz)) setzen und ADC aktivieren 
  ADMUX = mux_adc_channel;                      // bergebenen Kanal waehlen
// ADMUX |= (1<<REFS1) | (1<<REFS0); // interne Referenzspannung nutzen 
  ADMUX |=  (1<<REFS0); // VCC als Referenzspannung nutzen 
 
  /* nach Aktivieren des ADC wird ein "Dummy-Readout" empfohlen, man liest
     also einen Wert und verwirft diesen, um den ADC "warmlaufen zu lassen" */
  ADCSRA |= (1<<ADSC);              // eine ADC-Wandlung (Der ADC setzt dieses Bit ja wieder auf 0 nach dem Wandeln)
  while ( ADCSRA & (1<<ADSC) ) {
     ;     // auf Abschluss der Wandlung warten 
  }
 
  // Eigentliche Messung - Mittelwert aus 4 aufeinanderfolgenden Wandlungen
  for(i=0;i<4;i++)
  {
    ADCSRA |= (1<<ADSC);            // eine Wandlung
    while ( ADCSRA & (1<<ADSC) ) {
      ;     // auf Abschluss der Wandlung warten 
    }
    result += ADCW;            // Wandlungsergebnisse aufaddieren
  }
  ADCSRA &= ~(1<<ADEN);             // ADC deaktivieren ("Enable-Bit" auf LOW setzen)
 
  result /= 4;                     // Summe durch vier teilen = arithm. Mittelwert
 
  return result;
}

float PI_Regler(float ek_0)	/* Funktion PI-Regler*/
{

	static volatile float yk_0,yk_1;			/* yk = y(k), yk_1 = y(k-1) */
	static volatile float ek_1;				/* ek_1 = e(k-1), ek_2 = e(k-2)*/
	float b0 =  0.05706;
	float b1 = -0.00594;
	float a1 = -1;
	
	yk_0 = -1*a1*yk_1 + b0*ek + b1*ek_1;		/* PI-Regelalorithmus */
	
	yk_1 = yk_0;		/* Bereite Variablenwerte fr den nchsten */
	ek_1 = ek_0;		/* Unterprogrammaufruf vor */
	
	return yk_0;		/* Gib yk an das Hauptprogramm zurck */
}
